package ach.image;

import java.io.*;
import java.net.URL;
import java.awt.*;
import java.awt.image.ImageProducer;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import ach.image.EMemImageSource;
import ach.image.bmp.BmpImageProducer;
import ach.image.tiff.TiffImageProducer;
import ach.image.png.PngToolkit;
import ach.image.ImgProdObserver;
import ach.types.Rational;
import ach.vectorGraphics.*;


/**
 * ImageInfo.java                                                     <br>
 * Copyright (c) 1997, 2005 Helge Hackbarth, All Rights Reserved.     <br>
 * A class describing properties of an image according to tags defined
 * in the TIFF file format.                                           <br>
 * NOTE: Most of the image information fields like width, height,
 * xResolution, yResolution, maxGrayLevel, bitsPerSample etc. are only
 * set for TIFF files. If you need this information also for images
 * read from GIF or JPEG files, please refer to the methods / property
 * fields available through the EMemImageSource-class in this package.<br>
 *                                                                    <br>
 * THE AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. THE AUTHOR SHALL NOT BE LIABLE FOR ANY
 * DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * @author Helge Hackbarth
 * @version 1.2
 */
public class ImageInfo implements Runnable
{
  /**
   * Predefined image file format
   */
  public static final byte
    ifUNKNOWN = 0,
    ifGIF     = 1,
    ifJPG     = 2,
    ifBMP     = 3,
    ifTIF     = 4,
    ifPNG     = 5;
    // remember to update imageNameMasks and detectImageFormat
    // if more formats are added

  char psc = java.io.File.pathSeparatorChar;
  /**
   * Name-masks for file dialog (seperated by pathSeparatorChar)
   */
  public String imageNameMasks =
    "*.bmp" + psc + "*.gif" + psc + "*.jpg" + psc + "*.jpeg" + psc +
    "*.tif" + psc + "*.tiff" + psc + "*.fob" + psc + "*.eps" + psc + "*.png";

  /**
   * Filename extension that is expected for annotation file (default: ".xml")
   */
  public String annotationExtension = ".xml";

  /**
   * Sub-format for image data
   */
  public static final byte
    sfUNKNOWN = 0,
    INTEL     = 1,
    MOTOROLA  = 2,
    GIF87     = 3,
    GIF89A    = 4;

  /**
   * Compression type
   */
  public static final int
    UNCOMPRESSED = 1,
    CCITTG3      = 2,
    CCITTFAXG3   = 3,
    CCITTFAXG4   = 4,
    LZW          = 5,
    JPEG         = 6,
    PACKBITS     = 32773;

  /**
   * Photometric interpretation
   */
  public static final byte
    GRAYWHITE0   = 0,
    GRAYBLACK0   = 1,
    RGB          = 2,
    PALETTECOL   = 3,
    TRANSPARMASK = 4,
    CMYK         = 5,
    YCBCR        = 6;

  /**
   * Index for RGB-Palette
   */
  public static final byte
    REDIDX     = 0,
    GREENIDX   = 1,
    BLUEIDX    = 2;

  /**
   * (Path-)name of image file
   */
  public String imgName = "";
  /**
   * URL of image file (read from local file if URL == null)
   */
  public URL imgURL = null;
  /**
   * ByteArray that might contain the image-file as byte array
   * (alternative to passing the file as imgName or imgURL)
   */
  public byte[] imgFileBA = null;
  /**
   * Current image file format
   */
  public byte imgFormat = ifUNKNOWN;
  /**
   * Current image file subformat (set to ifGIF / ifTIF / ...).
   */
  public byte subFormat = sfUNKNOWN;
  /**
   * Number of pages. TIFF files might contain more than one page.
   */
  public int numPages = 1;
  /**
   * Current of page number within a multi document TIFF file (counting start at 1).
   */
  public int page = 1;
  /**
   * Flag if image was modified
   */
  public boolean hasChanged = false;
  /**
   * Horizontal number of pixels
   */
  public int width = 0;
  /**
   * Vertical number of pixels
   */
  public int height = 0;
  /**
   * Horizontal resolution in pixels per inch
   */
  public Rational xResolution = new Rational(96,1);
  /**
   * Vertical resolution in pixels per inch
   */
  public Rational yResolution = new Rational(96,1);
  /**
   * Maximum gray level (e.g. 15 for 4 bit images, 255 for 8 bit images)
   */
  public short maxGrayLevel = 1;
  /**
   * Number of bytes per row (if stored bitwise packed)
   */
  public int bytesPerRow = 0;
  // the following vars must be set to default values for each new image:
  /**
   * Bits per sample
   * @see #samplesPerPixel
   */
  public short bitsPerSample = 1;
  /**
   * Samples per pixel (e.g. 3 for RGB true color images)
   */
  public short samplesPerPixel = 1;
  /**
   * Number of sample planes (in rare cases red, green and blue
   * samples might be stored in seperate planes).
   */
  public short samplePlanes = 1;
  /**
   * Current compression type
   */
  public int compression = UNCOMPRESSED;
  /**
   * Predictor controls whether the image data are horizontally differentiated
   */
  public short predictor = 1;
  /**
   * Current photometric interpretation
   */
  public short photometricInt = 0;
  /**
   * Sample values continously or seperated in planes
   * @see #samplePlanes
   */
  public short planarConfig = 1;
  /**
   * Gray response unit (controls interpretation of gray response curve)
   */
  public short grayRespUnit = 3;
  /**
   * Gray response curve (allows a nonlinear gray distribution)
   */
  public short[] grayRespCurve = new short[256];
  /**
   * RGB-palette (array of 256 * 3 bytes for red, green and blue.
   * 0 <= red, green, blue <= 255.
   */
  public byte[] rgbPalette = new byte[256 * (BLUEIDX + 1)];
  /**
   * Artist that created that image
   */
  public String artist = "";
  /**
   * Software that was used to create that image
   */
  public String software = "";
  /**
   * Description of the image
   */
  public String imageDescription = "";
  /**
   * Flag if annotations are allowed to be read for the image (static, so it controls all instances)
   */
  public static boolean annotationsAllowed = true;
  /**
   * Annotations (XML formatted string) for the image
   */
  public String annotations = "";
  /**
   * Array of ShapeLists that contains annotations parsed into vector graphic objects.
   * The array contains one Shapelist for each page if ImageInfo represents a multipage file.
   */
  public ShapeList[] shapeLists = null;
  /**
   * Orientation of the image
   * orientation values:
   * 1 NOP                                                           <br>
   * 2 mirror horizontally                                           <br>
   * 3 rotate 180 degree                                             <br>
   * 4 rotate 180 degree and mirror horizontally                     <br>
   * 5 rotate 270 degree anticlockwise and mirror horizontally       <br>
   * 6  rotate 270 degree anticlockwise                              <br>
   * 7  rotate 90 degree anticlockwise and mirror horizontally       <br>
   * 8  rotate 90 degree anticlockwise
   */
  public short orientation = 1;

  /**
   * Current Java style ColorModel
   */
  public ColorModel colorModel = null;
  /**
   * Array of bytes containing image data (if palette image with more than 16 colors)
   */
  public byte bytePixels[] = null;
  /**
   * Array of integers containing image data (if truecolor image)
   */
  public int intPixels[] = null;
  /**
   * Array of bytes containing packed image data (if palette image with 16 or less colors)
   */
  public byte packedPixels[] = null;
  /**
   * Error message of last operation
   */
  public String errorMsg = null;

  private boolean versionEqualGreater1_1 =
    (System.getProperty("java.version").compareTo("1.1") >= 0);
  private boolean versionEqualGreater1_2 =
    (System.getProperty("java.version").compareTo("1.2") >= 0);
  private Toolkit toolkit = null;
  /**
   * imageProducer created by method createImageProducer
   */
  private ImageProducer imageProducer = null;
  private ImgProdObserver imgProdObserver = null;
  private boolean asynchronous = true; // how to run createProducer

  /**
   *  Constructor for ImageInfo
   */
  public ImageInfo()
  {
    super();
  }

  /**
   * Method to detect image format from file extension
   */
  public void detectImageFormat()
  {
    if (imgName == null) return;
    String imgExt = "";
    int index; // Make last part of URL  the name
    if ((index = imgName.lastIndexOf('.')) >= 0)
      imgExt = imgName.substring(index);
    if (imgExt.equalsIgnoreCase(".bmp"))
      imgFormat = ifBMP;
    else if (imgExt.equalsIgnoreCase(".gif"))
      imgFormat = ifGIF;
    else if (imgExt.equalsIgnoreCase(".jpg") ||
         imgExt.equalsIgnoreCase(".jpeg"))
      imgFormat = ifJPG;
    else if (imgExt.equalsIgnoreCase(".png"))
      imgFormat = ifPNG;
    else if (imgExt.equalsIgnoreCase(".tif") ||
         imgExt.equalsIgnoreCase(".tiff") ||
         imgExt.equalsIgnoreCase(".fob") || // fob used by FileNet
         imgExt.equalsIgnoreCase(".eps"))
      imgFormat = ifTIF;
    else
      imgFormat = ifTIF; // assume Tiff
  }

  /**
   * Set default values for important image info variables
   */
  public void setDefaults()
  {
    imgName = "";
    imgURL = null;
    imgFormat = ifUNKNOWN;
    subFormat = sfUNKNOWN;
    hasChanged = false;    // Flag, ob Bild geaendert wurde
    width = 0;
    height = 0;
    xResolution = new Rational(96,1);
    yResolution = new Rational(96,1);
    maxGrayLevel = 1;
    bytesPerRow = 0;
    bitsPerSample = 1;
    samplesPerPixel = 1;
    samplePlanes = 1;
    compression = UNCOMPRESSED;
    predictor = 1;
    photometricInt = 1;
    planarConfig = 1; // Sample values jedes Pixel kontinuierlich oder in Planes
    grayRespUnit = 3;
    artist = "";
    software = "";
    imageDescription = "";
    annotations = "";
    orientation = 1;
    colorModel = null;
    bytePixels = null;
    intPixels = null;
    packedPixels = null;
    page = 1;
    numPages = 1;
  }

  /**
   * Load annotation file (expected with same name like imgName or imgUrl but
   * with extension ".xml").
   */
  public void getAnnotations() {
    String xmlFileName = imgName;
    if (imgURL != null)
      xmlFileName = imgURL.getFile();
    int li = xmlFileName.lastIndexOf(".");
    if (li >=0)
      xmlFileName = xmlFileName.substring(0,li);
    xmlFileName = xmlFileName + annotationExtension;
    BufferedInputStream br;
    try {
      if (imgURL != null) {
        URL xmlURL = new URL(imgURL.getProtocol(), imgURL.getHost(), imgURL.getPort(), xmlFileName);
        br = new BufferedInputStream(xmlURL.openStream());
      } else {
        br = new BufferedInputStream(new FileInputStream(xmlFileName));
      }
      int c;
      StringBuffer sb = new StringBuffer();
      // detect if file looks like a XML formatted file (search for '<' at the beginning)
      while ((c = br.read()) != -1)
        if (((char)c != ' ') && ((char)c != '\r') && ((char)c != '\n')) break;
      if (c != -1) {
        if ((char)c == '<') {
          sb.append((char)c);
          while ((c = br.read()) != -1)
            sb.append((char)c);
        }
        else
          System.err.println("Annotations not XML-formatted");
      }
      br.close();
      annotations = sb.toString();
      shapeLists = new Annotations().createShapeLists(annotations);
      for (int i=0; i<shapeLists.length; i++)
        shapeLists[i].resetModifiedState();
    } catch (Exception e) { annotations = null; }
  }

  /**
   * Create an image producer from imgName or imgUrl.
   * Requires that Toolkit was set before (using method setAsynchronousTools())
   */
  public void createImageProducer()
    throws AWTException
  {
    int count = 3; // retry count
    errorMsg = null;
    detectImageFormat();
    if (imgURL == null) {
      if (imgName == null) return;
      try { // try to interpret name as URL first
        imgURL = new URL(imgName);
      } catch (Exception e) { // now try as file if URL creation failed
        while (count > 0) {
          try {
            File f = new File(imgName);
            imgName = f.getAbsolutePath(); // otherwise path gets lost if multipage images are printed to PDf
            if (f.canRead())
              count = 0;
            else {
              try { Thread.sleep(1000); } catch (InterruptedException ie) { }
              count--;
            }
          } catch (Exception e2) {
            try { Thread.sleep(1000); } catch (InterruptedException ie) { }
            count--;
          }
        }
      }
    }
    try {
      switch (imgFormat) {
        case ifBMP:
          if (asynchronous) { // create an ImageProducer for asynchronous loading
            imageProducer = new BmpImageProducer(imgURL, imgName);
          } else { // create an EMemImageSource as ImageProducer
            if (imgURL != null)
              imageProducer = (EMemImageSource)BmpImageProducer.getImageProducer(imgURL);
            else
              imageProducer = (EMemImageSource)BmpImageProducer.getImageProducer(imgName);
          }
          break;

        case ifTIF:
          if ((imgURL != null) && (imgFileBA == null))
            imgFileBA = ach.file.ParamUtil.netFileToRam(imgURL, !asynchronous);
          if (asynchronous) { // create an ImageProducer for asynchronous loading
            if (imgFileBA != null)
              imageProducer = new TiffImageProducer(imgFileBA, page);
            else
              imageProducer = new TiffImageProducer(imgURL, imgName, page);
          } else { // create an EMemImageSource as ImageProducer
            if (imgFileBA != null)
              imageProducer = (EMemImageSource)TiffImageProducer.getImageProducer(imgFileBA, page);
            else
              imageProducer = (EMemImageSource)TiffImageProducer.getImageProducer(imgURL, imgName, page);
            numPages = ((EMemImageSource)imageProducer).numPages;
            page = ((EMemImageSource)imageProducer).page;
          }
          break;

        case ifPNG:
          if (!versionEqualGreater1_1)
            throw new AWTException("This Java Virtual Machine does not support ZIP compression for PNG format");
          if (imgURL != null)
            imageProducer = PngToolkit.getImageProducer(imgURL);
          else
            imageProducer = PngToolkit.getImageProducer(imgName);
          break;

        default:
          if (toolkit == null) {
            errorMsg = "Toolkit not set";
            break;
          }
          if (imgFileBA != null)
            imageProducer = toolkit.createImage(imgFileBA).getSource();
          else {
            // Depending on the version of the underlying Java virtual machine, we will get
            // an ImageProducer now either using toolkit.createImage(...)
            // or toolkit.getImage(...).
            if (!versionEqualGreater1_2) { // JDK 1.1 does not support createImage(String)
              if (imgURL == null)
                imageProducer = toolkit.getImage(imgName).getSource();
              else
                imageProducer = toolkit.getImage(imgURL).getSource();
            }
            else {
              // execution had to be moved into a different class because otherwise the
              // precompilation of Microsofts JIT Compiler would throw an exception for
              // a non existing method although this method would never be reached
              imageProducer = EMemImageSource.getToolkitImageProducer_VerGreater12(toolkit, imgName, imgURL);
            }
          }
          break;
      }
    }
    catch (OutOfMemoryError oe) {
      errorMsg = oe.toString() + " other Java VM recommended - please view the readme file.";
    }
    catch (AWTException ae) {
      final String searchStr = "Exception: ";
      errorMsg = ae.toString();
      int li = errorMsg.lastIndexOf(searchStr);
      if (li >= 0)
        errorMsg = errorMsg.substring(li + searchStr.length());
    }
    catch (Throwable te) {
      errorMsg = te.toString();
    }
    // before reading new annotations, delete current if exist
    annotations = "";
    if (shapeLists != null) {
      for (int i=0; i < shapeLists.length; i++)
        shapeLists[i] = null;
      shapeLists = null;
    }
    if (annotationsAllowed)
      getAnnotations();
    if (imgProdObserver != null)
      imgProdObserver.imageProducerUpdate(imageProducer, errorMsg, this);
    if ((errorMsg != null) && (errorMsg.length() > 0))
      throw new AWTException(errorMsg);
  }

  /**
   * Allows to run the creation of an image producer asynchronously
   * as seperate thread.
   * Requires that Toolkit was set before (use method setAsynchronousTools())
   */
  public void run()
  {
    try {
      createImageProducer();
    }
    catch (AWTException ae)
    { }
  }

  /**
   * Define all necessary parameters to be able to start
   * image production as an asynchronous thread afterwards.
   * @param toolkit is the Toolkit provided by AWT classes
   * @param imgProdObserver used during asynchronous loading to
   * inform the observing class when loading is finished.
   * @param mode controls whether loading is done asynchronously or synchronously
   */
  public void setAsynchronousTools(Toolkit toolkit, ImgProdObserver imgProdObserver, boolean asyncMode) {
    this.toolkit = toolkit;
    this.imgProdObserver = imgProdObserver;
    this.asynchronous = asyncMode;
  }


  /**
   * calulate real number of bits per pixel according to mapSize of ColorModel
   * Note: ColorModel.getPixelSize() usually reports 8 even if mapsize < 256
   * (especially required for JDK 1.2)
   * @return (packed) bits per pixel according to mapSize of ColorModel
  **/
  public int calcPixelSize() {
    if (colorModel instanceof IndexColorModel) {
      int colors = ((IndexColorModel)colorModel).getMapSize();
      int pixelSize = 0;
      while (colors > 1) {
        pixelSize++;
        colors /= 2;
      }
      return pixelSize;
    }
    else
      return colorModel.getPixelSize();
  }

  /**
   * Returns information about the image
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer("");

    sb.append("Name: " + imgName);
    sb.append("\r\nURL: " + imgURL);
    sb.append("\r\nFormat: " + imgFormat);
    sb.append("\r\nSubFormat: ");
    switch (subFormat) {
    case INTEL:
      sb.append("Intel");
      break;
    case MOTOROLA:
      sb.append("Motorola");
      break;
    case GIF87:
      sb.append("GIF87");
      break;
    case GIF89A:
      sb.append("GIF89a");
      break;
    default:
      sb.append(subFormat);
      break;
    }
    sb.append("\r\nChanged: " + hasChanged);
    sb.append("\r\nWidth: " + width);
    sb.append("\r\nHeight: " + height);
    sb.append("\r\nX-Resolution: " + xResolution.numerator + " / " + xResolution.denominator + " dpi");
    sb.append("\r\nY-Resolution: " + yResolution.numerator + " / " + yResolution.denominator + " dpi");
    sb.append("\r\nMax. Palette-Index / Graylevel: " + maxGrayLevel);
    sb.append("\r\nBytes per Row: " + bytesPerRow);
    sb.append("\r\nBits per Sample: " + bitsPerSample);
    sb.append("\r\nSamples per Pixel: " + samplesPerPixel);
    sb.append("\r\nSample planes: " + samplePlanes);
    sb.append("\r\nCompression: ");
    switch (compression) {
      case UNCOMPRESSED:
        sb.append("uncompressed");
        break;
      case CCITTG3:
        sb.append("CCITT Group 3");
        break;
      case CCITTFAXG3:
        sb.append("CCITT Fax Group 3");
        break;
      case CCITTFAXG4:
        sb.append("CCITT Fax Group 4");
        break;
      case LZW:
        sb.append("LZW");
        break;
      case JPEG:
        sb.append("JPEG");
        break;
      case PACKBITS:
        sb.append("Packbits");
        break;
      default:
        sb.append(compression);
    }
    sb.append("\r\nPredictor: " + predictor);
    sb.append("\r\nPhotometric interpretation: ");
    switch (photometricInt) {
      case GRAYWHITE0:
        sb.append("Graylevel (white=0)");
        break;
      case GRAYBLACK0:
        sb.append("Graylevel (black=0)");
        break;
      case RGB:
        sb.append("RGB true color");
        break;
      case PALETTECOL:
        sb.append("Color palette");
        break;
      case TRANSPARMASK:
        sb.append("Transparency mask");
        break;
      case CMYK:
        sb.append("CMYK true color");
        break;
      case YCBCR:
        sb.append("YcBcR");
        break;
      default:
        sb.append(photometricInt);
    }
    sb.append("\r\nPlanar configuration: " + planarConfig);
    sb.append("\r\nArtist: " + artist);
    sb.append("\r\nSoftware: " + software);
    sb.append("\r\nImage description: " + imageDescription);
    sb.append("\r\nOrientation: " + orientation);
    return sb.toString();
  }

}

